import java.lang.StringBuffer;
import java.util.ArrayList;
import java.util.Arrays;

public class CrcProvider {
	//!/bin/env python
	// A library to work with CRC
	//  (c) 2005 - Intrepid Software
	//   Bas Westerbaan <bas.westerbaan@gmail.com>

	// Note on reflecting:
	//  Reflecting is pretty simple, reflecting the polynomial is just good enough.

    ArrayList table;
    int width;
    long basePolynomial;
    int reflected;
    long initial;
    long xorOut;
    long polynomial;
    int swapped;
    long _hash;


    CrcProvider() {    	
    }
    
    CrcProvider(int Width, long Polynomial, int Reflected, long Initial, long XorOut, int Swapped) {
		this.width=Width;
		this.basePolynomial = Polynomial;
		this.reflected = Reflected;
		this.initial = Initial;
		this.xorOut = XorOut;
		this.swapped = Swapped;
		
		if (this.reflected > 0)
			this.polynomial = ReflectNumber(this.basePolynomial, this.width * 8);
		else
			this.polynomial=this.basePolynomial;
		
		table = this.generateTable();
		this.reset();
	}

    // show crc table in hexadecimal
    public void showTable() {
    	Datamanipulation dm = new Datamanipulation();
    	
    	String r=new String();
        if (this.table.size() > 0)
            r = "" + dm.PadStringLeft(dm.NumberToHexadecimal(Long.valueOf(this.table.get(0).toString()).longValue()), this.width * 2, '0'); 
        else
            r = "";
        for (int i=1; i<table.size(); i++)
            r += ", " + dm.PadStringLeft(dm.NumberToHexadecimal(Long.valueOf(this.table.get(i).toString()).longValue()), this.width * 2, '0');
        r += "";
        System.out.println(r);
    }

    // generates a crc table
    public ArrayList generateTable() {
    	Datamanipulation dm = new Datamanipulation();
    	
        ArrayList ret = new ArrayList();
        long poly = ReflectNumber(this.basePolynomial, 32);
        for (int i=0; i<256; i++) {
            long entry = i;
            for (int j=8; j>0; j--) {
                if ((entry & 1) > 0){
                    entry = dm.Xor((entry >> 1), poly, this.width);
                }
                else
                    entry >>= 1;
            }
            if (this.reflected > 0)
                ret.add(Long.valueOf(entry));
            else
                ret.add(Long.valueOf(ReflectNumber(entry)));
        }
        return ret;
        }

    public void reset() {
    	this._hash = this.initial;
    }
    
    public long getHash() {
    	long h;
    	
        if (this.swapped > 0)
            h = this.crcByteSwap(this._hash);
        else
            h = this._hash;
        
        return h ^ this.xorOut;
    }

    // does a crc byteswap as in the flhash
    public long crcByteSwap(long hash) {
    	Datamanipulation dm = new Datamanipulation();
    	long h;
    	
        if (this.width != 4){
            System.out.println("Crc32 only supports byteswap");
//            throw new Exception("Crc32 only supports byteswap");
        }
        if (this.reflected >0){
            System.out.println("byteswap is a dirty reflection, use one of them");
//            throw new Exception("byteswap is a dirty reflection, use one of them");
        }
        h = (hash >> 24) | ((hash >> 8) & 0x0000ff00) | ((hash << 24) & 4278190080L) | ((hash << 8) & 0x00ff0000);

        return (h >> 2) | dm.NumberFromHexadecimal("80000000");
    }
    
    // update the hash with a string of data
    public void update(String s) {
    	Datamanipulation dm = new Datamanipulation();
        for (int i=0; i<s.length(); i++)
            this._hash = dm.Xor(this._hash >> 8, Long.valueOf(this.table.get((int)dm.Xor((int)s.toCharArray()[i], this._hash & 0xff, 1)).toString()).longValue(), this.width);
        }

    // calculates the bytes neccessary to append to get from the current
    //  to the specified hash.
    // NB: SWAPPED ISN'T SUPPORTED
    // There are some commented out debug print's, which can be very
    //     usefull.
    // this algorithm is vaguely derived from Anarchriz's method.
    public String patch(long wanted) {
    	Datamanipulation dm = new Datamanipulation();
        long w = dm.Xor(wanted, this.xorOut, this.width);            
        ArrayList lut = new ArrayList();
        long[] a = new long[table.size()];
        for (int i=0; i<table.size(); i++)
            a[i]=Long.valueOf(table.get(i).toString()).longValue();
//      lut.sort();
        Arrays.sort(a);
        for (int i=0; i<table.size(); i++)
            lut.add(Long.toString(a[i]));
        ArrayList m = new ArrayList();
        String r = "";
        for (int p=0; p<this.width; p++)
            m.add(Long.valueOf((this._hash & ((long)Math.pow(256, (p+1))-1)) >> (8 * p)));
        for (int p=0; p<this.width; p++)
            m.add(Long.valueOf((w & ((long)Math.pow(256, (p+1))-1)) >> (8 * p)));
        for (int p=this.width-1; p>=0; p--) {
        	long o = Long.valueOf(lut.get(Integer.valueOf(m.get(p+this.width).toString()).intValue()).toString()).longValue();
            //pm(m)
            //System.out.println("s: " + hex(m[p+self.width]) + " v: " + dm.NumberToHexadecimal(o) + " i: " + hex(this.table.get(o));
            //for (int db=0; db<p; db++)
            //	System.out.println(Long.toHexString(Long.valueOf(m.get(db).toString()).longValue()));
            for (int q=0; q<this.width; q++){
                long v = (o & ((long)Math.pow(256, (q+1))-1)) >> (8 * q);
            //#    print hex(v) + " ^ " + hex(m[p+q+1]) + ' -> ' + hex(Xor(m[p+q+1],v,1))
                m.set(p+q+1, Long.valueOf(dm.Xor(Long.valueOf(m.get(p+q+1).toString()).longValue(),v,1)));
            }    
            m.set(p, Long.valueOf(dm.Xor(Long.valueOf(m.get(p).toString()).longValue(), Long.valueOf(this.table.indexOf(Long.valueOf(o))).longValue(),1)));
            //#for db in range(p+self.width,self.width*2-1):
            //#    print hex(m[db])
        }
        //pm(m)
        for (int p=0; p<this.width; p++)
            r += String.valueOf((char)Integer.valueOf(m.get(p).toString()).intValue());
        return r;
    }
    
    //this.hash = this.getHash();

//    public void pm(String m) {
//    	Datamanipulation dm=new Datamanipulation();
//        String r = "";
        
//        for (int i=0; i<m.length(); i++)
//        	r += " " + dm.PadStringLeft(dm.NumberToHexadecimal(m.charAt(i)), 2, '0');
//        System.out.println(r);
//    }
    
    // gets the binary representation of the number and turns it around.
	//  algorithm could be more efficient.
	// this function can be used to get a reflected polynomial to generate
	//  a reversed crc table.
	// The default width is 4 bytes, this is the extra '0' padding done to ensure 1
	//  will become 2147483648 instead of 1.
    public long ReflectNumber(long number){
    	return ReflectNumber(number, 32);
    }

	public long ReflectNumber(long number, int width) {
		Datamanipulation dm = new Datamanipulation();
		String b = dm.PadStringLeft(dm.NumberToBinary(number), width, '0');
		return (long)dm.NumberFromBinary(new StringBuffer(b).reverse().toString());
	}
}
